use std::path::Path;

use serde::{Deserialize, Serialize};
use serde_json::Value;
use thiserror::Error;
use tokio::fs;
use tracing::{error, warn};

use crate::string::ToUnquotedString;

#[derive(Debug, Error)]
pub enum CommentsError {
    #[error("Target video {0} not found in database")]
    InvalidTarget(String),
    #[error("IO Error: {0}")]
    IOError(#[from] std::io::Error),
    #[error("Could not serialize info JSON: {0}")]
    SerializeInfoJSON(#[from] serde_json::Error),
}

#[allow(clippy::struct_excessive_bools)]
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Comment {
    pub id: String,
    pub video_id: String,
    pub parent: String,
    pub text: String,
    pub like_count: i64,
    pub author_id: String,
    pub author: String,
    pub author_thumbnail: String,
    pub author_url: String,
    pub timestamp: i64,
    pub author_is_uploader: bool,
    pub author_is_verified: bool,
    pub is_hearted: bool,
    pub is_pinned: bool,
}

pub async fn get_comments_from_video(id: &str) -> Result<Vec<Comment>, CommentsError> {
    let info_json = format!("media/videos/{id}/{id}.info.json");
    let info_json = Path::new(&info_json);

    if !info_json.exists() {
        return Err(CommentsError::InvalidTarget(id.into()));
    }

    let info: Value = serde_json::from_str(&fs::read_to_string(info_json).await?)?;

    let Some(comments) = info.get("comments") else {
        warn!("Video {id} has no comments!");
        return Ok(vec![]);
    };

    comments.as_array().map_or_else(
        || Ok(vec![]),
        |comments| {
            Ok(comments
                .iter()
                .map(|c| Comment {
                    id: c["id"].to_unquoted_string(),
                    video_id: id.into(),
                    parent: c["parent"].to_unquoted_string(),
                    text: c["text"].to_unquoted_string(),
                    like_count: c["like_count"].as_i64().unwrap_or_default(),
                    author_id: c["author_id"].to_unquoted_string(),
                    author: c["author"].to_unquoted_string(),
                    author_thumbnail: c["author_thumbnail"].to_unquoted_string(),
                    author_url: c["author_url"].to_unquoted_string(),
                    timestamp: c["timestamp"].as_i64().unwrap_or_default(),
                    author_is_uploader: c["author_is_uploader"].as_bool().unwrap_or_default(),
                    author_is_verified: c["author_is_verified"].as_bool().unwrap_or_default(),
                    is_hearted: c["is_favorited"].as_bool().unwrap_or_default(),
                    is_pinned: c["is_pinned"].as_bool().unwrap_or_default(),
                })
                .collect::<Vec<_>>())
        },
    )
}
