use axum::{
    Json,
    extract::{Path, Query, State},
    http::StatusCode,
};
use serde::{Deserialize, Serialize};
use tracing::{error, info};
use url::Url;

use crate::{
    channel::{Channel, ChannelError},
    instance::Instance,
    url::is_youtube_url,
};

#[derive(Debug, Deserialize)]
pub struct ListChannelsQuery {
    page: Option<usize>,
}

#[derive(Debug, Serialize)]
pub struct ListChannelsResponse {
    channels: Vec<Channel>,
    page: usize,
    per_page: usize,
    total: usize,
    pages: usize,
}

/// Retrieve video list as JSON (paged)
pub async fn list_channels(
    State(state): State<Instance>,
    Query(query): Query<ListChannelsQuery>,
) -> Result<Json<ListChannelsResponse>, StatusCode> {
    let Ok(channels) = sqlx::query_as!(Channel, "SELECT * FROM channel")
        .fetch_all(&state.pool)
        .await
    else {
        error!("Could not fetch channels from database!");
        return Err(StatusCode::INTERNAL_SERVER_ERROR);
    };

    let per_page = state.config.pagination.channels;
    let total = channels.len();
    let pages = total.div_ceil(per_page);
    let page = query.page.unwrap_or(1).min(pages).max(1);

    let start = per_page * (page - 1);
    let end = (start + per_page).min(total);

    let channels = if start < total {
        channels[start..end].to_vec()
    } else {
        vec![]
    };

    Ok(Json(ListChannelsResponse {
        channels,
        page,
        per_page,
        total,
        pages,
    }))
}

/// Get a single channel from the database by its ID
pub async fn get_channel(
    State(state): State<Instance>,
    Path(id): Path<String>,
) -> Result<Json<Channel>, StatusCode> {
    sqlx::query_as!(Channel, "SELECT * FROM channel WHERE youtube_id = ?", id)
        .fetch_optional(&state.pool)
        .await
        .map_or(Err(StatusCode::INTERNAL_SERVER_ERROR), |channel| {
            channel.map_or(Err(StatusCode::NOT_FOUND), |c| Ok(Json(c)))
        })
}

#[derive(Debug, Deserialize)]
pub struct UploadChannelQuery {
    url: String,
}

/// Upload a channel's metadata to the database
pub async fn upload_channel(
    State(state): State<Instance>,
    Query(query): Query<UploadChannelQuery>,
) -> StatusCode {
    let id = match sqlx::query_scalar!("SELECT MAX(id) FROM channel")
        .fetch_one(&state.pool)
        .await
    {
        Ok(Some(max_id)) => max_id + 1,
        Ok(None) => 0,
        Err(_) => {
            return StatusCode::INTERNAL_SERVER_ERROR;
        }
    };

    let Ok(url) = Url::parse(&query.url) else {
        error!("Could not parse URL!");
        return StatusCode::BAD_REQUEST;
    };

    if !is_youtube_url(&url) {
        error!("YouTube URL RegEx match failed!");
        return StatusCode::BAD_REQUEST;
    }

    let new_channel = Channel::from_url(&url, id).await.map_err(|e| match e {
        ChannelError::AlreadyExists => StatusCode::OK,
        _ => StatusCode::INTERNAL_SERVER_ERROR,
    });

    match new_channel {
        Ok(channel) => {
            match sqlx::query!(
                "
                INSERT INTO channel (
                    id, url, youtube_id, name, handle_url, avatar_url, 
                    banner_url, description, subscribers
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
                ",
                channel.id,
                channel.url,
                channel.youtube_id,
                channel.name,
                channel.handle_url,
                channel.avatar_url,
                channel.banner_url,
                channel.description,
                channel.subscribers,
            )
            .execute(&state.pool)
            .await
            {
                Ok(result) => {
                    info!("Inserted channel to database successfully! {result:?}");
                    StatusCode::OK
                }
                Err(e) => {
                    error!("Error inserting channel to database: {e:?}");
                    StatusCode::INTERNAL_SERVER_ERROR
                }
            }
        }
        Err(status) => status,
    }
}
