use axum::{
    Json,
    extract::{Path, Query, State},
    http::StatusCode,
};
use serde::{Deserialize, Serialize};
use tracing::error;

use crate::{
    comment::{Comment, CommentsError, get_comments_from_video},
    instance::Instance,
};

#[derive(Debug, Deserialize)]
pub struct VideoCommentsQuery {
    page: Option<usize>,
}

#[derive(Debug, Serialize)]
pub struct VideoCommentsResponse {
    comments: Vec<Comment>,
    page: usize,
    per_page: usize,
    total: usize,
    pages: usize,
}

/// Fetches the comments from a video, will return an empty vec if the video has no comments
pub async fn get_comments(
    State(state): State<Instance>,
    Path(id): Path<String>,
    Query(query): Query<VideoCommentsQuery>,
) -> Result<Json<VideoCommentsResponse>, StatusCode> {
    let comments = get_comments_from_video(&id).await.map_err(|e| match e {
        CommentsError::InvalidTarget(t) => {
            error!("Video {t} does not exist in database!");
            StatusCode::NOT_FOUND
        }
        _ => StatusCode::INTERNAL_SERVER_ERROR,
    });

    match comments {
        Ok(comments) => {
            let per_page = state.config.pagination.comments;
            let total = comments.len();
            let pages = total.div_ceil(per_page);
            let page = query.page.unwrap_or(1).max(1).min(pages);

            let start = per_page * (page - 1);
            let end = (start + per_page).min(total);

            let comments = if start < total {
                comments[start..end].to_vec()
            } else {
                vec![]
            };

            Ok(Json(VideoCommentsResponse {
                comments,
                page,
                per_page,
                total,
                pages,
            }))
        }
        Err(status) => Err(status),
    }
}
