use std::net::SocketAddr;

use axum::{
    extract::{ConnectInfo, Request, State},
    http::StatusCode,
    middleware::Next,
    response::Response,
};
use serde::Deserialize;
use sha3::{Digest, Sha3_256};
use tracing::error;

use crate::instance::Instance;

#[derive(Debug, PartialEq, Eq, Deserialize)]
pub struct Key(pub String);

#[derive(Debug)]
pub enum KeyError {
    Empty,
    Invalid,
}

impl Key {
    /// Checks if the API key is valid by testing it against a hash.
    fn validate(&self, state: &Instance) -> Result<(), KeyError> {
        let k = &self.0;
        if k.is_empty() {
            return Err(KeyError::Empty);
        }

        let hash = state.key_hash.clone();

        if format!("{:x}", Sha3_256::digest(k.as_bytes())) == hash {
            Ok(())
        } else {
            Err(KeyError::Invalid)
        }
    }
}
pub async fn auth(
    ConnectInfo(addr): ConnectInfo<SocketAddr>,
    State(state): State<Instance>,
    req: Request,
    next: Next,
) -> Result<Response, StatusCode> {
    let Some(auth_header) = req
        .headers()
        .get("almond-api-key")
        .and_then(|h| h.to_str().ok())
    else {
        error!("Could not find almond-api-key header (from {addr:?})");
        return Err(StatusCode::UNAUTHORIZED);
    };

    let key = Key(auth_header.into());

    if matches!(key.validate(&state), Ok(())) {
        Ok(next.run(req).await)
    } else {
        error!("Bad password for request (from {addr:?})");
        Err(StatusCode::UNAUTHORIZED)
    }
}
